@@ -22,7 +22,6 @@
  */
 
 #include "config.h"
-#include "galias.h"
 
 #include "gkeyfile.h"
 
@@ -61,6 +60,8 @@
 #include "gstrfuncs.h"
 #include "gutils.h"
 
+#include "galias.h"
+
 typedef struct _GKeyFileGroup GKeyFileGroup;
 
 struct _GKeyFile
@@ -342,6 +343,8 @@ find_file_in_data_dirs (const gchar   *file,
   if (output_file != NULL && fd > 0)
     *output_file = g_strdup (path);
 
+  g_free (path);
+
   return fd;
 }
 
@@ -485,8 +488,9 @@ g_key_file_load_from_file (GKeyFile       *key_file,
  * @flags: flags from #GKeyFileFlags
  * @error: return location for a #GError, or %NULL
  *
- * Loads a key file from memory into an empty #GKeyFile structure.  
- * If the object cannot be created then %error is set to a #GKeyFileError.
+ * Loads a key file from memory into an empty #GKeyFile structure.  If
+ * the object cannot be created then %error is set to a
+ * #GKeyFileError. 
  *
  * Return value: %TRUE if a key file could be loaded, %FALSE othewise
  * Since: 2.6
@@ -504,6 +508,9 @@ g_key_file_load_from_data (GKeyFile       *key_file,
   g_return_val_if_fail (data != NULL, FALSE);
   g_return_val_if_fail (length != 0, FALSE);
 
+  if (length == (gsize)-1)
+    length = strlen (data);
+
   if (key_file->approximate_size > 0)
     {
       g_key_file_clear (key_file);
@@ -847,6 +854,11 @@ g_key_file_parse_data (GKeyFile     *key_file,
     {
       if (data[i] == '\n')
         {
+	  if (i > 0 && data[i - 1] == '\r')
+	    g_string_erase (key_file->parse_buffer,
+			    key_file->parse_buffer->len - 1,
+			    1);
+	    
           /* When a newline is encountered flush the parse buffer so that the
            * line can be parsed.  Note that completely blank lines won't show
            * up in the parse buffer, so they get parsed directly.
@@ -1007,16 +1019,16 @@ g_key_file_get_keys (GKeyFile     *key_file,
   keys = (gchar **) g_new0 (gchar **, num_keys + 1);
 
   tmp = group->key_value_pairs;
-  for (i = 0; i < num_keys; i++)
+  for (i = 1; i <= num_keys; i++)
     {
       GKeyFileKeyValuePair *pair;
 
       pair = (GKeyFileKeyValuePair *) tmp->data;
-      keys[i] = g_strdup (pair->key);
+      keys[num_keys - i] = g_strdup (pair->key);
 
       tmp = tmp->next;
     }
-  keys[i] = NULL;
+  keys[num_keys] = NULL;
 
   if (length)
     *length = num_keys;
@@ -1118,8 +1130,8 @@ g_key_file_get_groups (GKeyFile *key_file,
  * event that the @group_name cannot be found, %NULL is returned 
  * and @error is set to #G_KEY_FILE_ERROR_GROUP_NOT_FOUND.
  *
- * Return value: a string or %NULL if the specified key cannot be
- * found.
+ * Return value: a newly allocated string or %NULL if the specified 
+ * key cannot be found.
  *
  * Since: 2.6
  **/
@@ -1224,8 +1236,8 @@ g_key_file_set_value (GKeyFile    *key_file,
  * event that the @group_name cannot be found, %NULL is returned 
  * and @error is set to #G_KEY_FILE_ERROR_GROUP_NOT_FOUND.
  *
- * Return value: a string or %NULL if the specified key cannot be
- * found.
+ * Return value: a newly allocated string or %NULL if the specified 
+ * key cannot be found.
  *
  * Since: 2.6
  **/
@@ -1506,8 +1518,9 @@ extern GSList *_g_compute_locale_variants (const gchar *locale);
  * with @key cannot be interpreted or no suitable translation can
  * be found then the untranslated value is returned.
  *
- * Return value: a string or %NULL if the specified key cannot be
- *               found.
+ * Return value: a newly allocated string or %NULL if the specified key 
+ *  cannot be found.
+ * 
  * Since: 2.6
  **/
 gchar *
@@ -1595,8 +1608,7 @@ g_key_file_get_locale_string (GKeyFile     *key_file,
  *
  * Returns the values associated with @key under @group_name
  * translated in the given @locale if available.  If @locale is
- * %NULL then the current locale is assumed. If @group_name is
- * %NULL, then the start group is used.
+ * %NULL then the current locale is assumed. 
 
  * If @key cannot be found then %NULL is returned and @error is set to
  * #G_KEY_FILE_ERROR_KEY_NOT_FOUND. If the values associated
@@ -1661,8 +1673,7 @@ g_key_file_get_locale_string_list (GKeyFile     *key_file,
  *
  * Associates a list of string values for @key and @locale under
  * @group_name.  If the translation for @key cannot be found then
- * it is created. If @group_name is %NULL, the start group is
- * used.
+ * it is created. 
  *
  * Since: 2.6
  **/
@@ -1711,7 +1722,7 @@ g_key_file_set_locale_string_list (GKeyFile            *key_file,
  * @error: return location for a #GError
  *
  * Returns the value associated with @key under @group_name as a
- * boolean. If @group_name is %NULL, the start group is used.
+ * boolean. 
  *
  * If @key cannot be found then the return value is undefined and
  * @error is set to #G_KEY_FILE_ERROR_KEY_NOT_FOUND. Likewise, if
@@ -1776,8 +1787,7 @@ g_key_file_get_boolean (GKeyFile     *key_file,
  * @value: %TRUE or %FALSE
  *
  * Associates a new boolean value with @key under @group_name.
- * If @key cannot be found then it is created. If @group_name
- * is %NULL, the start group is used.
+ * If @key cannot be found then it is created. 
  *
  * Since: 2.6
  **/
@@ -1832,7 +1842,7 @@ g_key_file_get_boolean_list (GKeyFile     *key_file,
   gsize i, num_bools;
 
   g_return_val_if_fail (key_file != NULL, NULL);
-  g_return_val_if_fail (group_name != NULL, FALSE);
+  g_return_val_if_fail (group_name != NULL, NULL);
   g_return_val_if_fail (key != NULL, NULL);
 
   key_file_error = NULL;
@@ -1993,8 +2003,7 @@ g_key_file_get_integer (GKeyFile     *key_file,
  * @value: an integer value
  *
  * Associates a new integer value with @key under @group_name.
- * If @key cannot be found then it is created. If @group_name
- * is %NULL, the start group is used.
+ * If @key cannot be found then it is created. 
  *
  * Since: 2.6
  **/
@@ -2096,7 +2105,6 @@ g_key_file_get_integer_list (GKeyFile     *key_file,
  *
  * Associates a list of integer values with @key under
  * @group_name.  If @key cannot be found then it is created.
- * If @group_name is %NULL the start group is used.
  *
  * Since: 2.6
  **/
@@ -2557,8 +2565,7 @@ g_key_file_has_group (GKeyFile    *key_file,
  * @error: return location for a #GError
  *
  * Looks whether the key file has the key @key in the group
- * @group_name. If @group_name is %NULL, the start group is
- * used.
+ * @group_name. 
  *
  * Return value: %TRUE if @key is a part of @group_name, %FALSE
  * otherwise.
@@ -2755,12 +2762,15 @@ g_key_file_remove_group (GKeyFile     *key_file,
   group_node = g_key_file_lookup_group_node (key_file, group_name);
 
   if (!group_node)
-    g_set_error (error, G_KEY_FILE_ERROR,
-		 G_KEY_FILE_ERROR_GROUP_NOT_FOUND,
-		 _("Key file does not have group '%s'"),
-		 group_name);
+    {
+      g_set_error (error, G_KEY_FILE_ERROR,
+		   G_KEY_FILE_ERROR_GROUP_NOT_FOUND,
+		   _("Key file does not have group '%s'"),
+		   group_name);
+      return;
+    }
 
-  g_key_file_remove_group_node (key_file, group_node);
+    g_key_file_remove_group_node (key_file, group_node);
 }
 
 static void
@@ -2817,7 +2827,6 @@ g_key_file_remove_key (GKeyFile     *key_file,
       return;
     }
 
-  group->key_value_pairs = g_list_remove (group->key_value_pairs, key_file);
   pair = g_key_file_lookup_key_value_pair (key_file, group, key);
 
   if (!pair)
@@ -2829,9 +2838,10 @@ g_key_file_remove_key (GKeyFile     *key_file,
       return;
     }
 
-  g_hash_table_remove (group->lookup_map, pair->key);
-
   key_file->approximate_size -= strlen (pair->key) + strlen (pair->value) + 2;
+
+  group->key_value_pairs = g_list_remove (group->key_value_pairs, pair);
+  g_hash_table_remove (group->lookup_map, pair->key);  
   g_key_file_key_value_pair_free (pair);
 }
 
@@ -2967,7 +2977,6 @@ g_key_file_parse_value_as_string (GKeyFile     *key_file,
 				  GSList      **pieces,
 				  GError      **error)
 {
-  GError *parse_error = NULL;
   gchar *string_value, *p, *q0, *q;
 
   string_value = g_new0 (gchar, strlen (value) + 1);
@@ -3002,6 +3011,13 @@ g_key_file_parse_value_as_string (GKeyFile     *key_file,
               *q = '\\';
               break;
 
+	    case '\0':
+	      g_set_error (error, G_KEY_FILE_ERROR,
+			   G_KEY_FILE_ERROR_INVALID_VALUE,
+			   _("Key file contains escape character "
+			     "at end of line"));
+	      break;
+
             default:
 	      if (pieces && *p == key_file->list_separator)
 		*q = key_file->list_separator;
@@ -3010,7 +3026,7 @@ g_key_file_parse_value_as_string (GKeyFile     *key_file,
 		  *q++ = '\\';
 		  *q = *p;
 		  
-		  if (parse_error == NULL)
+		  if (*error == NULL)
 		    {
 		      gchar sequence[3];
 		      
@@ -3037,15 +3053,13 @@ g_key_file_parse_value_as_string (GKeyFile     *key_file,
 	    }
 	}
 
+      if (*p == '\0')
+	break;
+
       q++;
       p++;
     }
 
-  if (p > value && p[-1] == '\\' && error == NULL)
-    g_set_error (error, G_KEY_FILE_ERROR,
-		 G_KEY_FILE_ERROR_INVALID_VALUE,
-		 _("Key file contains escape character at end of line"));
-
   *q = '\0';
   if (pieces)
   {
@@ -3153,14 +3167,28 @@ g_key_file_parse_value_as_integer (GKeyFile     *key_file,
 				   GError      **error)
 {
   gchar *end_of_valid_int;
-  gint int_value = 0;
+  glong long_value;
+  gint int_value;
 
-  int_value = strtol (value, &end_of_valid_int, 0);
+  errno = 0;
+  long_value = strtol (value, &end_of_valid_int, 10);
 
-  if (*end_of_valid_int != '\0')
-    g_set_error (error, G_KEY_FILE_ERROR,
-		 G_KEY_FILE_ERROR_INVALID_VALUE,
-		 _("Value '%s' cannot be interpreted as a number."), value);
+  if (*value == '\0' || *end_of_valid_int != '\0')
+    {
+      g_set_error (error, G_KEY_FILE_ERROR,
+		   G_KEY_FILE_ERROR_INVALID_VALUE,
+		   _("Value '%s' cannot be interpreted as a number."), value);
+      return 0;
+    }
+
+  int_value = long_value;
+  if (int_value != long_value || errno == ERANGE)
+    {
+      g_set_error (error, G_KEY_FILE_ERROR,
+		   G_KEY_FILE_ERROR_INVALID_VALUE,
+		   _("Integer value '%s' out of range"), value);
+      return 0;
+    }
 
   return int_value;
 }
@@ -3255,4 +3283,5 @@ g_key_file_parse_comment_as_value (GKeyFile      *key_file,
   return value;
 }
 
-
+#define __G_KEY_FILE_C__
+#include "galiasdef.c"
