@@ -105,6 +105,8 @@ const guint glib_micro_version = GLIB_MICRO_VERSION;
 const guint glib_interface_age = GLIB_INTERFACE_AGE;
 const guint glib_binary_age = GLIB_BINARY_AGE;
 
+G_WIN32_DLLMAIN_FOR_DLL_NAME (static, dll_name)
+
 /**
  * glib_check_version:
  * @required_major: the required major version.
@@ -1295,6 +1297,26 @@ get_special_folder (int csidl)
   return retval;
 }
 
+static char *
+get_windows_directory_root (void)
+{
+  char windowsdir[MAX_PATH];
+
+  if (GetWindowsDirectory (windowsdir, sizeof (windowsdir)))
+    {
+      /* Usually X:\Windows, but in terminal server environments
+       * might be an UNC path, AFAIK.
+       */
+      char *p = (char *) g_path_skip_root (windowsdir);
+      if (G_IS_DIR_SEPARATOR (p[-1]) && p[-2] != ':')
+	p--;
+      *p = '\0';
+      return g_strdup (windowsdir);
+    }
+  else
+    return g_strdup ("C:\\");
+}
+
 #endif
 
 /* HOLDS: g_utils_global_lock */
@@ -1309,6 +1331,10 @@ g_get_any_init (void)
       if (!g_tmp_dir)
 	g_tmp_dir = g_strdup (g_getenv ("TEMP"));
       
+#ifdef G_OS_WIN32
+      if (!g_tmp_dir)
+	g_tmp_dir = get_windows_directory_root ();
+#else
 #ifdef P_tmpdir
       if (!g_tmp_dir)
 	{
@@ -1322,12 +1348,9 @@ g_get_any_init (void)
       
       if (!g_tmp_dir)
 	{
-#ifndef G_OS_WIN32
 	  g_tmp_dir = g_strdup ("/tmp");
-#else /* G_OS_WIN32 */
-	  g_tmp_dir = g_strdup ("\\");
-#endif /* G_OS_WIN32 */
 	}
+#endif	/* !G_OS_WIN32 */
       
 #ifdef G_OS_WIN32
       /* We check $HOME first for Win32, though it is a last resort for Unix
@@ -1367,17 +1390,7 @@ g_get_any_init (void)
 	g_home_dir = get_special_folder (CSIDL_PROFILE);
       
       if (!g_home_dir)
-	{
-	  /* At least at some time, HOMEDRIVE and HOMEPATH were used
-	   * to point to the home directory, I think. But on Windows
-	   * 2000 HOMEDRIVE seems to be equal to SYSTEMDRIVE, and
-	   * HOMEPATH is its root "\"?
-	   */
-	  if (g_getenv ("HOMEDRIVE") != NULL && g_getenv ("HOMEPATH") != NULL)
-	    g_home_dir = g_strconcat (g_getenv ("HOMEDRIVE"),
-				      g_getenv ("HOMEPATH"),
-				      NULL);
-	}
+	g_home_dir = get_windows_directory_root ();
 #endif /* G_OS_WIN32 */
       
 #ifdef HAVE_PWD_H
@@ -1896,7 +1909,11 @@ g_get_system_data_dirs (void)
   if (!g_system_data_dirs)
     {
 #ifdef G_OS_WIN32
+      gchar *glib_top_share_dir, *exe_top_share_dir;
+
+      /* Documents and Settings\All Users\Application Data */
       char *appdata = get_special_folder (CSIDL_COMMON_APPDATA);
+      /* Documents and Settings\All Users\Documents */
       char *docs = get_special_folder (CSIDL_COMMON_DOCUMENTS);
       
       if (appdata && docs)
@@ -1913,18 +1930,48 @@ g_get_system_data_dirs (void)
       else if (docs)
 	data_dirs = docs;
       else
-	data_dirs = NULL;
+	data_dirs = g_strdup ("");
 
-      if (data_dirs)
+      /* Using the above subfolders of Documents and Settings perhaps
+       * makes sense from a Windows perspective.
+       *
+       * But looking at the actual use cases of this function in GTK+
+       * and GNOME software, what we really want is the "share"
+       * subdirectory of the installation directory for the package
+       * our caller is a part of.
+       *
+       * As we don't know who calls us, punt, and use the installation
+       * location of GLib, and of the .exe file being run. To guard
+       * against neither of those being what we really want, callers
+       * of this function should have Win32-specific code to look up
+       * their installation folder themselves, and handle a subfolder
+       * "share" of it in the same way as the folders returned from
+       * this function.
+       */
+      glib_top_share_dir = g_win32_get_package_installation_subdirectory (NULL, dll_name, "share");
+
+      if (glib_top_share_dir)
 	{
-	  data_dir_vector = g_strsplit (data_dirs, G_SEARCHPATH_SEPARATOR_S, 0);
-	  g_free (data_dirs);
+	  gchar *tem = data_dirs;
+	  data_dirs = g_strconcat (data_dirs, G_SEARCHPATH_SEPARATOR_S,
+				   glib_top_share_dir, NULL);
+	  g_free (tem);
+	  g_free (glib_top_share_dir);
 	}
-      else
+
+      exe_top_share_dir = g_win32_get_package_installation_subdirectory (NULL, NULL, "share");
+
+      if (exe_top_share_dir)
 	{
-	  /* Punt, return empty list */
-	  data_dir_vector = g_strsplit ("", G_SEARCHPATH_SEPARATOR_S, 0);
+	  gchar *tem = data_dirs;
+	  data_dirs = g_strconcat (data_dirs, G_SEARCHPATH_SEPARATOR_S,
+				   exe_top_share_dir, NULL);
+	  g_free (tem);
+	  g_free (exe_top_share_dir);
 	}
+
+      data_dir_vector = g_strsplit (data_dirs, G_SEARCHPATH_SEPARATOR_S, 0);
+      g_free (data_dirs);
 #else
       data_dirs = (gchar *) g_getenv ("XDG_DATA_DIRS");
 
@@ -1997,6 +2044,8 @@ g_get_system_config_dirs (void)
   return (G_CONST_RETURN gchar * G_CONST_RETURN *) conf_dir_vector;
 }
 
+#ifndef G_OS_WIN32
+
 static GHashTable *alias_table = NULL;
 
 /* read an alias file for the locales */
@@ -2052,9 +2101,12 @@ read_aliases (gchar *file)
   fclose (fp);
 }
 
+#endif
+
 static char *
 unalias_lang (char *lang)
 {
+#ifndef G_OS_WIN32
   char *p;
   int i;
 
@@ -2075,6 +2127,7 @@ unalias_lang (char *lang)
 	  return lang;
 	}
     }
+#endif
   return lang;
 }
 
@@ -2394,25 +2447,72 @@ _g_utils_thread_init (void)
 
 #include <libintl.h>
 
-#ifdef G_PLATFORM_WIN32
-
-G_WIN32_DLLMAIN_FOR_DLL_NAME (static, dll_name)
+#ifdef G_OS_WIN32
 
+/**
+ * _glib_get_locale_dir:
+ *
+ * Return the path to the lib\locale subfolder of the GLib
+ * installation folder. The path is in the system codepage. We have to
+ * use system codepage as bindtextdomain() doesn't have a UTF-8
+ * interface.
+ */
 static const gchar *
 _glib_get_locale_dir (void)
 {
-  static const gchar *cache = NULL;
-  if (cache == NULL)
-    cache = g_win32_get_package_installation_subdirectory
-      (GETTEXT_PACKAGE, dll_name, "lib\\locale");
+  gchar *dir;
+  gchar *cp_dir;
+  gchar *retval = NULL;
 
-  return cache;
+  dir = g_win32_get_package_installation_directory (GETTEXT_PACKAGE, dll_name);
+  if (dir != NULL)
+    {
+      /* Convert to system codepage. We *must* pass a NULL error
+       * pointer so that the functions that g_locale_from_utf8() calls
+       * don't attempt to store a translated error message into the
+       * error in case of failure. Note that these functions have been
+       * specially modified to not call g_set_error() if the error
+       * pointer is NULL. The translation would cause a recursive
+       * invocation of _glib_gettext(), and infinite recursion.
+       */
+      cp_dir = g_locale_from_utf8 (dir, -1, NULL, NULL, NULL);
+
+      if (cp_dir == NULL && G_WIN32_HAVE_WIDECHAR_API ())
+	{
+	  /* It failed, so convert to wide chars, check if there is a
+	   * 8.3 version and use that.
+	   */
+	  wchar_t *wdir = g_utf8_to_utf16 (dir, -1, NULL, NULL, NULL);
+	  if (wdir != NULL)
+	    {
+	      wchar_t wshortdir[MAX_PATH + 1];
+	      if (GetShortPathNameW (wdir, wshortdir, G_N_ELEMENTS (wshortdir)))
+		{
+		  g_free (dir);
+		  dir = g_utf16_to_utf8 (wshortdir, -1, NULL, NULL, NULL);
+		  cp_dir = g_locale_from_utf8 (dir, -1, NULL, NULL, NULL);
+		}
+	      g_free (wdir);
+	    }
+	}
+
+      if (cp_dir != NULL)
+	retval = g_strconcat (cp_dir, "\\lib\\locale", NULL);
+      
+      g_free (dir);
+      g_free (cp_dir);
+    }
+
+  if (retval)
+    return retval;
+  else
+    return g_strdup ("");
 }
 
 #undef GLIB_LOCALE_DIR
 #define GLIB_LOCALE_DIR _glib_get_locale_dir ()
 
-#endif /* G_PLATFORM_WIN32 */
+#endif /* G_OS_WIN32 */
 
 G_CONST_RETURN gchar *
 _glib_gettext (const gchar *str)
